@@ -27,7 +27,8 @@ class Restraint(object):
     """Base class for all restraints.
        See :attr:`ihm.System.restraints`.
     """
-    pass
+    def _get_report(self):
+        return str(self)
 
 
 class RestraintGroup(list):
@@ -58,7 +59,8 @@ class EM3DRestraint(Restraint):
        :param assembly: The part of the system that is fit into the map.
        :type assembly: :class:`~ihm.Assembly`
        :param bool segment: True iff the map has been segmented.
-       :param str fitting_method: The method used to fit the model into the map.
+       :param str fitting_method: The method used to fit the model
+              into the map.
        :param fitting_method_citation: The publication describing the fitting
               method.
        :type fitting_method_citation: :class:`~ihm.Citation`
@@ -67,6 +69,12 @@ class EM3DRestraint(Restraint):
        :param str details: Additional details regarding the fitting.
     """
 
+    def _get_report(self):
+        ret = "Fit to 3D electron microscopy density map"
+        if self.fitting_method:
+            ret += " using " + self.fitting_method
+        return ret
+
     def __init__(self, dataset, assembly, segment=None, fitting_method=None,
                  fitting_method_citation=None, number_of_gaussians=None,
                  details=None):
@@ -89,7 +97,7 @@ class EM3DRestraintFit(object):
        :param float cross_correlation_coefficient: The fit between the model
               and the map.
     """
-    __slots__ = ["cross_correlation_coefficient"] # Reduce memory usage
+    __slots__ = ["cross_correlation_coefficient"]  # Reduce memory usage
 
     def __init__(self, cross_correlation_coefficient=None):
         self.cross_correlation_coefficient = cross_correlation_coefficient
@@ -114,6 +122,13 @@ class SASRestraint(Restraint):
        :param str details: Additional details regarding the fitting.
     """
 
+    def _get_report(self):
+        state_map = {True: "Multi-state ", False: "Single-state "}
+        ret = "%sSAS restraint" % state_map.get(self.multi_state, "")
+        if self.fitting_atom_type:
+            ret += " on " + self.fitting_atom_type
+        return ret
+
     def __init__(self, dataset, assembly, segment=None, fitting_method=None,
                  fitting_atom_type=None, multi_state=None,
                  radius_of_gyration=None, details=None):
@@ -136,7 +151,7 @@ class SASRestraintFit(object):
 
        :param float chi_value: The fit between the model and the SAS data.
     """
-    __slots__ = ["chi_value"] # Reduce memory usage
+    __slots__ = ["chi_value"]  # Reduce memory usage
 
     def __init__(self, chi_value=None):
         self.chi_value = chi_value
@@ -165,6 +180,9 @@ class EM2DRestraint(Restraint):
        :param str details: Additional details regarding the fitting.
     """
 
+    def _get_report(self):
+        return "Fit to 2D electron microscopy class average"
+
     def __init__(self, dataset, assembly, segment=None,
                  number_raw_micrographs=None, pixel_size_width=None,
                  pixel_size_height=None, image_resolution=None,
@@ -196,7 +214,7 @@ class EM2DRestraintFit(object):
               places the model on the image.
     """
     __slots__ = ["cross_correlation_coefficient",
-                 "rot_matrix", "tr_vector"] # Reduce memory usage
+                 "rot_matrix", "tr_vector"]  # Reduce memory usage
 
     def __init__(self, cross_correlation_coefficient=None,
                  rot_matrix=None, tr_vector=None):
@@ -214,7 +232,12 @@ class CrossLinkRestraint(Restraint):
        :type linker: :class:`ihm.ChemDescriptor`
     """
 
-    assembly = None # no struct_assembly_id for XL restraints
+    assembly = None  # no struct_assembly_id for XL restraints
+
+    def _get_report(self):
+        return ("%d %s cross-links from %d experimental identifications"
+                % (len(self.cross_links), self.linker.auth_name,
+                   sum(len(x) for x in self.experimental_cross_links)))
 
     def __init__(self, dataset, linker):
         self.dataset, self.linker = dataset, linker
@@ -236,6 +259,9 @@ class CrossLinkRestraint(Restraint):
 class ExperimentalCrossLink(object):
     """A cross-link identified in the experiment.
 
+       These objects, once created, should be added to
+       the :attr:`CrossLinkRestraint.experimental_cross_links` list.
+
        :param residue1: The first residue linked by the cross-link.
        :type residue1: :class:`ihm.Residue`
        :param residue2: The second residue linked by the cross-link.
@@ -248,7 +274,7 @@ class ExperimentalCrossLink(object):
 
 
 class DistanceRestraint(object):
-    """Abstract base class for all distance restraints. These are typically
+    """Base class for all distance restraints. These are typically
        used in a :class:`DerivedDistanceRestraint`.
 
        Do not use this class directly but instead use a derived class
@@ -257,7 +283,15 @@ class DistanceRestraint(object):
        :class:`LowerBoundDistanceRestraint`,
        or :class:`LowerUpperBoundDistanceRestraint`.
     """
-    pass
+    restraint_type = None
+
+    #: The minimum distance allowed for this restraint,
+    #: or None if unconstrained
+    distance_lower_limit = None
+
+    #: The maximum distance allowed for this restraint,
+    #: or None if unconstrained
+    distance_upper_limit = None
 
 
 class HarmonicDistanceRestraint(DistanceRestraint):
@@ -267,10 +301,14 @@ class HarmonicDistanceRestraint(DistanceRestraint):
        :param float distance: Equilibrium distance
     """
     restraint_type = 'harmonic'
+
     def __init__(self, distance):
         self.distance = distance
 
+    #: The equilibrium distance
     distance_lower_limit = property(lambda self: self.distance)
+
+    #: The equilibrium distance
     distance_upper_limit = distance_lower_limit
 
 
@@ -281,10 +319,14 @@ class UpperBoundDistanceRestraint(DistanceRestraint):
        :param float distance: Distance threshold
     """
     restraint_type = 'upper bound'
+
     def __init__(self, distance):
         self.distance = distance
 
+    #: The maximum distance allowed by this restraint
     distance_upper_limit = property(lambda self: self.distance)
+
+    #: Minimum distance (unconstrained, so always None)
     distance_lower_limit = None
 
 
@@ -295,10 +337,14 @@ class LowerBoundDistanceRestraint(DistanceRestraint):
        :param float distance: Distance threshold
     """
     restraint_type = 'lower bound'
+
     def __init__(self, distance):
         self.distance = distance
 
+    #: The minimum distance allowed by this restraint
     distance_lower_limit = property(lambda self: self.distance)
+
+    #: Maximum distance (unconstrained, so always None)
     distance_upper_limit = None
 
 
@@ -311,8 +357,12 @@ class LowerUpperBoundDistanceRestraint(DistanceRestraint):
        :param float distance_upper_limit: Upper bound on the distance.
     """
     restraint_type = 'lower and upper bound'
+
     def __init__(self, distance_lower_limit, distance_upper_limit):
+        #: The minimum distance allowed by this restraint
         self.distance_lower_limit = distance_lower_limit
+
+        #: The maximum distance allowed by this restraint
         self.distance_upper_limit = distance_upper_limit
 
 
@@ -345,7 +395,11 @@ class CrossLinkPseudoSite(object):
 
 
 class ResidueCrossLink(CrossLink):
-    """A cross-link used in the modeling, applied to residue alpha carbon atoms.
+    """A cross-link used in the modeling, applied to residue
+       alpha carbon atoms.
+
+       These objects, once created, should be added to
+       the :attr:`CrossLinkRestraint.cross_links` list.
 
        :param experimental_cross_link: The corresponding cross-link identified
               by experiment. Multiple cross-links can map to a single
@@ -382,16 +436,33 @@ class ResidueCrossLink(CrossLink):
         self.distance, self.restrain_all = distance, restrain_all
         self.pseudo1, self.pseudo2 = pseudo1, pseudo2
 
-        #: Information about the fit of each model to this cross-link
+        #: Information about the fit of each model to this cross-link.
         #: This is a Python dict where keys are :class:`~ihm.model.Model`
         #: objects and values are :class:`CrossLinkFit` objects.
         self.fits = {}
 
+    def _get_residue1(self):
+        seq_id = self.experimental_cross_link.residue1.seq_id
+        return self.asym1.residue(seq_id)
+    residue1 = property(_get_residue1,
+                        doc="Residue object representing one end "
+                            "of the cross-link")
+
+    def _get_residue2(self):
+        seq_id = self.experimental_cross_link.residue2.seq_id
+        return self.asym2.residue(seq_id)
+    residue2 = property(_get_residue2,
+                        doc="Residue object representing one end "
+                            "of the cross-link")
+
 
 class FeatureCrossLink(CrossLink):
     """A cross-link used in the modeling, applied to the closest primitive
        object with the highest resolution.
 
+       These objects, once created, should be added to
+       the :attr:`CrossLinkRestraint.cross_links` list.
+
        :param experimental_cross_link: The corresponding cross-link identified
               by experiment. Multiple cross-links can map to a single
               experimental identification.
@@ -427,7 +498,7 @@ class FeatureCrossLink(CrossLink):
         self.distance, self.restrain_all = distance, restrain_all
         self.pseudo1, self.pseudo2 = pseudo1, pseudo2
 
-        #: Information about the fit of each model to this cross-link
+        #: Information about the fit of each model to this cross-link.
         #: This is a Python dict where keys are :class:`~ihm.model.Model`
         #: objects and values are :class:`CrossLinkFit` objects.
         self.fits = {}
@@ -436,6 +507,9 @@ class FeatureCrossLink(CrossLink):
 class AtomCrossLink(CrossLink):
     """A cross-link used in the modeling, applied to the specified atoms.
 
+       These objects, once created, should be added to
+       the :attr:`CrossLinkRestraint.cross_links` list.
+
        :param experimental_cross_link: The corresponding cross-link identified
               by experiment. Multiple cross-links can map to a single
               experimental identification.
@@ -473,7 +547,7 @@ class AtomCrossLink(CrossLink):
         self.distance, self.restrain_all = distance, restrain_all
         self.pseudo1, self.pseudo2 = pseudo1, pseudo2
 
-        #: Information about the fit of each model to this cross-link
+        #: Information about the fit of each model to this cross-link.
         #: This is a Python dict where keys are :class:`~ihm.model.Model`
         #: objects and values are :class:`CrossLinkFit` objects.
         self.fits = {}
@@ -488,7 +562,7 @@ class CrossLinkFit(object):
        :param float sigma1: Uncertainty in the position of the first residue.
        :param float sigma2: Uncertainty in the position of the second residue.
     """
-    __slots__ = ["psi", "sigma1", "sigma2"] # Reduce memory usage
+    __slots__ = ["psi", "sigma1", "sigma2"]  # Reduce memory usage
 
     def __init__(self, psi=None, sigma1=None, sigma2=None):
         self.psi, self.sigma1, self.sigma2 = psi, sigma1, sigma2
@@ -504,6 +578,7 @@ class Feature(object):
        :class:`~ihm.restraint.DerivedDistanceRestraint` objects.
     """
     details = None
+
     def _all_entities_or_asyms(self):
         # Get all Entities or AsymUnits referenced by this object
         return []
@@ -512,14 +587,14 @@ class Feature(object):
 class ResidueFeature(Feature):
     """Selection of one or more residues from the system.
 
-       Residues can be selected from both :class:`AsymUnit` and
-       :class:`Entity` (the latter implies that it selects residues
+       Residues can be selected from both :class:`ihm.AsymUnit` and
+       :class:`ihm.Entity` (the latter implies that it selects residues
        in all instances of that entity). Individual residues can
-       also be selected by passing :class:`Residue` objects.
+       also be selected by passing :class:`ihm.Residue` objects.
 
-       :param sequence ranges: A list of :class:`AsymUnitRange`,
-              :class:`AsymUnit`, :class:`EntityRange`, :class:`Residue`,
-              and/or :class:`Entity` objects.
+       :param sequence ranges: A list of :class:`ihm.AsymUnitRange`,
+              :class:`ihm.AsymUnit`, :class:`ihm.EntityRange`,
+              :class:`ihm.Residue`, and/or :class:`ihm.Entity` objects.
        :param str details: Additional text describing this feature.
     """
 
@@ -541,7 +616,9 @@ class ResidueFeature(Feature):
 
     def _get_entity_type(self):
         def _get_entity(x):
-            return x if isinstance(x, ihm.Entity) else x.entity
+            if isinstance(x, ihm.Entity):
+                return x
+            return x.entity if x.entity else x.asym.entity
         if any(not _get_entity(r).is_polymeric() for r in self.ranges):
             raise ValueError("%s cannot select non-polymeric entities" % self)
         else:
@@ -551,8 +628,8 @@ class ResidueFeature(Feature):
 class AtomFeature(Feature):
     """Selection of one or more atoms from the system.
        Atoms can be selected from polymers or non-polymers (but not both).
-       Atoms can also be selected from both :class:`AsymUnit` and
-       :class:`Entity` (the latter implies that it selects atoms
+       Atoms can also be selected from both :class:`ihm.AsymUnit` and
+       :class:`ihm.Entity` (the latter implies that it selects atoms
        in all instances of that entity).
        For selecting an entire polymer or residue(s),
        see :class:`ResidueFeature`. For selecting an entire non-polymer,
@@ -582,12 +659,12 @@ class NonPolyFeature(Feature):
     """Selection of one or more non-polymers from the system.
        To select individual atoms from a non-polymer, see :class:`AtomFeature`.
 
-       Features can include both :class:`AsymUnit` and
-       :class:`Entity` (the latter implies that it selects non-polymers
+       Features can include both :class:`ihm.AsymUnit` and
+       :class:`ihm.Entity` (the latter implies that it selects non-polymers
        in all instances of that entity).
 
-       :param sequence objs: A list of :class:`AsymUnit` and/or
-              :class:`Entity` objects.
+       :param sequence objs: A list of :class:`ihm.AsymUnit` and/or
+              :class:`ihm.Entity` objects.
        :param str details: Additional text describing this feature.
     """
 
@@ -604,7 +681,8 @@ class NonPolyFeature(Feature):
         def _get_entity(x):
             return x if isinstance(x, ihm.Entity) else x.entity
         if any(_get_entity(r).is_polymeric() for r in self.objs):
-            raise ValueError("%s can only select non-polymeric entities" % self)
+            raise ValueError(
+                "%s can only select non-polymeric entities" % self)
         else:
             return _get_entity(self.objs[0]).type if self.objs else None
 
@@ -625,7 +703,7 @@ class PseudoSiteFeature(Feature):
         return 'other'
 
 
-class GeometricRestraint(object):
+class GeometricRestraint(Restraint):
     """A restraint between part of the system and some part of a
        geometric object. See :class:`CenterGeometricRestraint`,
        :class:`InnerSurfaceGeometricRestraint`,
@@ -644,7 +722,11 @@ class GeometricRestraint(object):
        :param bool restrain_all: If True, all distances are restrained.
     """
     object_characteristic = 'other'
-    assembly = None # no struct_assembly_id for geometric restraints
+    assembly = None  # no struct_assembly_id for geometric restraints
+
+    def _get_report(self):
+        return ("Distance (%s) to %s"
+                % (self.distance.restraint_type, self.geometric_object.type))
 
     def __init__(self, dataset, geometric_object, feature, distance,
                  harmonic_force_constant=None, restrain_all=None,
@@ -680,7 +762,7 @@ class OuterSurfaceGeometricRestraint(GeometricRestraint):
     object_characteristic = 'outer surface'
 
 
-class DerivedDistanceRestraint(object):
+class DerivedDistanceRestraint(Restraint):
     """A restraint between two parts of the system, derived from experimental
        data.
 
@@ -698,7 +780,7 @@ class DerivedDistanceRestraint(object):
        :param float mic_value: Value of the Maximal Information Coefficient
               (MIC) for this interaction, if applicable.
     """
-    assembly = None # no struct_assembly_id for derived distance restraints
+    assembly = None  # no struct_assembly_id for derived distance restraints
 
     def __init__(self, dataset, feature1, feature2, distance,
                  probability=None, restrain_all=None, mic_value=None):
@@ -710,7 +792,7 @@ class DerivedDistanceRestraint(object):
     _all_features = property(lambda self: (self.feature1, self.feature2))
 
 
-class PredictedContactRestraint(object):
+class PredictedContactRestraint(Restraint):
     """A predicted contact between two parts of the system, derived from
        various computational tools.
 
@@ -731,7 +813,7 @@ class PredictedContactRestraint(object):
        :type software: :class:`~ihm.Software`
 
     """
-    assembly = None # no struct_assembly_id for predicted contact restraints
+    assembly = None  # no struct_assembly_id for predicted contact restraints
 
     def __init__(self, dataset, resatom1, resatom2, distance,
                  by_residue, probability=None, software=None):
