@@ -1,11 +1,6 @@
 import utils
 import os
 import unittest
-import sys
-if sys.version_info[0] >= 3:
-    from io import StringIO
-else:
-    from io import BytesIO as StringIO
 try:
     import urllib.request as urllib2
 except ImportError:
@@ -19,6 +14,8 @@ import ihm.representation
 import ihm.model
 import ihm.source
 import ihm.flr
+import ihm.multi_state_scheme
+
 
 class Tests(unittest.TestCase):
     def test_system(self):
@@ -42,6 +39,17 @@ class Tests(unittest.TestCase):
         self.assertEqual(cc1, cc2)
         self.assertEqual(hash(cc1), hash(cc2))
         self.assertNotEqual(cc1, cc3)
+        cc4 = ihm.ChemComp(id='GLY', code='G', code_canonical='G',
+                           formula=ihm.unknown)
+        self.assertIsNone(cc4.formula_weight)
+
+    def test_chem_comp_id_5(self):
+        """Test new-style 5-character CCD IDs in ChemComp"""
+        cc1 = ihm.ChemComp(id='MYGLY', code='G', code_canonical='G')
+        self.assertEqual(cc1.id, 'MYGLY')
+        self.assertEqual(cc1.code, 'G')
+        self.assertEqual(cc1.code_canonical, 'G')
+        self.assertEqual(cc1.type, 'other')
 
     def test_chem_comp_weight(self):
         """Test ChemComp.formula_weight"""
@@ -52,11 +60,14 @@ class Tests(unittest.TestCase):
         cc = ihm.ChemComp('X', 'X', 'X', formula='C90H')
         self.assertRaises(ValueError, lambda x: x.formula_weight, cc)
         # Formula with unknown element
-        cc = ihm.ChemComp('X', 'X', 'X', formula='C5 Y')
+        cc = ihm.ChemComp('X', 'X', 'X', formula='C5 Es')
         self.assertIsNone(cc.formula_weight)
         # Formula with known elements and no charge
         cc = ihm.ChemComp('X', 'X', 'X', formula='C6 H12 P')
         self.assertAlmostEqual(cc.formula_weight, 115.136, delta=0.01)
+        # Formula with element 'X' (e.g. GLX, ASX)
+        cc = ihm.ChemComp('X', 'X', 'X', formula='C6 H12 P X2')
+        self.assertAlmostEqual(cc.formula_weight, 115.136, delta=0.01)
         # Formula with known elements and formal charge
         cc = ihm.ChemComp('X', 'X', 'X', formula='C6 H12 P 1')
         self.assertAlmostEqual(cc.formula_weight, 115.136, delta=0.01)
@@ -81,10 +92,36 @@ class Tests(unittest.TestCase):
         cc1 = ihm.DNAChemComp(id='DG', code='DG', code_canonical='G')
         self.assertEqual(cc1.type, 'DNA linking')
 
+    def test_sugar_chem_comp(self):
+        """Test SaccharideChemComp class and subclasses"""
+        cc = ihm.SaccharideChemComp('NAG')
+        self.assertEqual(cc.type, 'saccharide')
+
+        cc = ihm.LSaccharideChemComp('NAG')
+        self.assertEqual(cc.type, 'L-saccharide')
+
+        cc = ihm.DSaccharideChemComp('NAG')
+        self.assertEqual(cc.type, 'D-saccharide')
+
+        cc = ihm.LSaccharideAlphaChemComp('NAG')
+        self.assertEqual(cc.type, 'L-saccharide, alpha linking')
+
+        cc = ihm.LSaccharideBetaChemComp('NAG')
+        self.assertEqual(cc.type, 'L-saccharide, beta linking')
+
+        cc = ihm.DSaccharideAlphaChemComp('NAG')
+        self.assertEqual(cc.type, 'D-saccharide, alpha linking')
+
+        cc = ihm.DSaccharideBetaChemComp('NAG')
+        self.assertEqual(cc.type, 'D-saccharide, beta linking')
+
     def test_non_polymer_chem_comp(self):
         """Test NonPolymerChemComp class"""
         cc1 = ihm.NonPolymerChemComp('HEM')
         self.assertEqual(cc1.type, 'non-polymer')
+        self.assertEqual(cc1.code_canonical, 'X')
+        cc2 = ihm.NonPolymerChemComp('HEM', code_canonical='G')
+        self.assertEqual(cc2.code_canonical, 'G')
 
     def test_water_chem_comp(self):
         """Test WaterChemComp class"""
@@ -105,13 +142,19 @@ class Tests(unittest.TestCase):
         self.assertAlmostEqual(a._comps['M'].formula_weight, 149.211,
                                delta=0.01)
 
+        self.assertEqual(a._comps['Z'].id, 'GLX')
+        self.assertEqual(a._comps['Z'].name, "GLU/GLN AMBIGUOUS")
+        self.assertEqual(a._comps['Z'].formula, 'C5 H8 N O2 X2')
+        self.assertAlmostEqual(a._comps['Z'].formula_weight, 114.124,
+                               delta=0.01)
+
         a = ihm.LPeptideAlphabet()
         self.assertIn('MSE', a)
         self.assertNotIn('DG', a)
-        self.assertEqual(len(a.keys), 22)
-        self.assertEqual(len(a.values), 22)
+        self.assertEqual(len(a.keys), 25)
+        self.assertEqual(len(a.values), 25)
         self.assertEqual(sorted(a.keys)[0], 'A')
-        self.assertEqual(len(a.items), 22)
+        self.assertEqual(len(a.items), 25)
         item0 = sorted(a.items)[0]
         self.assertEqual(item0[0], 'A')
         self.assertEqual(item0[1].id, 'ALA')
@@ -126,22 +169,23 @@ class Tests(unittest.TestCase):
 
     def test_d_peptide_alphabet(self):
         """Test DPeptideAlphabet class"""
-        dcode_from_canon =  {'A': 'DAL', 'C': 'DCY', 'D': 'DAS', 'E': 'DGL',
-                             'F': 'DPN', 'H': 'DHI', 'I': 'DIL', 'K': 'DLY',
-                             'L': 'DLE', 'M': 'MED', 'N': 'DSG', 'P': 'DPR',
-                             'Q': 'DGN', 'R': 'DAR', 'S': 'DSN', 'T': 'DTH',
-                             'V': 'DVA', 'W': 'DTR', 'Y': 'DTY', 'G': 'G'}
-        d = ihm.DPeptideAlphabet
-        l = ihm.LPeptideAlphabet
+        dcode_from_canon = {'A': 'DAL', 'C': 'DCY', 'D': 'DAS', 'E': 'DGL',
+                            'F': 'DPN', 'H': 'DHI', 'I': 'DIL', 'K': 'DLY',
+                            'L': 'DLE', 'M': 'MED', 'N': 'DSG', 'P': 'DPR',
+                            'Q': 'DGN', 'R': 'DAR', 'S': 'DSN', 'T': 'DTH',
+                            'V': 'DVA', 'W': 'DTR', 'Y': 'DTY', 'G': 'G'}
+        da = ihm.DPeptideAlphabet
+        la = ihm.LPeptideAlphabet
         # Weights and formulae of all standard amino acids should be identical
         # between L- and D- forms (except for lysine, where the formal charge
         # differs between the two forms)
         for canon in 'ACDEFGHILMNPQRSTVWY':
             lcode = canon
             dcode = dcode_from_canon[canon]
-            self.assertEqual(d._comps[dcode].formula, l._comps[lcode].formula)
-            self.assertAlmostEqual(d._comps[dcode].formula_weight,
-                                   l._comps[lcode].formula_weight, delta=0.01)
+            self.assertEqual(da._comps[dcode].formula,
+                             la._comps[lcode].formula)
+            self.assertAlmostEqual(da._comps[dcode].formula_weight,
+                                   la._comps[lcode].formula_weight, delta=0.01)
 
     def test_rna_alphabet(self):
         """Test RNAAlphabet class"""
@@ -159,9 +203,10 @@ class Tests(unittest.TestCase):
 
     def test_chem_descriptor(self):
         """Test ChemDescriptor class"""
-        d1 = ihm.ChemDescriptor('EDC', chemical_name='test-EDC',
-                chem_comp_id='test-chem-comp',
-                common_name='test-common-EDC', smiles='CCN=C=NCCCN(C)C')
+        d1 = ihm.ChemDescriptor(
+            'EDC', chemical_name='test-EDC',
+            chem_comp_id='test-chem-comp',
+            common_name='test-common-EDC', smiles='CCN=C=NCCCN(C)C')
         self.assertEqual(d1.auth_name, 'EDC')
         self.assertEqual(d1.chem_comp_id, 'test-chem-comp')
         self.assertEqual(d1.chemical_name, 'test-EDC')
@@ -173,16 +218,25 @@ class Tests(unittest.TestCase):
     def test_entity(self):
         """Test Entity class"""
         e1 = ihm.Entity('AHCD', description='foo')
-        # Should compare identical if sequences are the same
+        # Should compare identical if sequences are the same, if not branched
         e2 = ihm.Entity('AHCD', description='bar')
         e3 = ihm.Entity('AHCDE', description='foo')
         heme = ihm.Entity([ihm.NonPolymerChemComp('HEM')])
+        sugar = ihm.Entity([ihm.SaccharideChemComp('NAG'),
+                            ihm.SaccharideChemComp('FUC')])
         self.assertEqual(e1, e2)
         self.assertNotEqual(e1, e3)
-        self.assertEqual(e1.seq_id_range, (1,4))
-        self.assertEqual(e3.seq_id_range, (1,5))
+        self.assertEqual(e1.seq_id_range, (1, 4))
+        self.assertEqual(e3.seq_id_range, (1, 5))
+        sugar2 = ihm.Entity([ihm.SaccharideChemComp('NAG'),
+                             ihm.SaccharideChemComp('FUC')])
+        # Branched entities never compare equal unless they are the same object
+        self.assertEqual(sugar, sugar)
+        self.assertNotEqual(sugar, sugar2)
         # seq_id does not exist for nonpolymers
-        self.assertEqual(heme.seq_id_range, (None,None))
+        self.assertEqual(heme.seq_id_range, (None, None))
+        # We do have an internal seq_id_range for branched entities
+        self.assertEqual(sugar.seq_id_range, (1, 2))
 
     def test_entity_weight(self):
         """Test Entity.formula_weight"""
@@ -197,12 +251,50 @@ class Tests(unittest.TestCase):
         protein = ihm.Entity('AHCD')
         heme = ihm.Entity([ihm.NonPolymerChemComp('HEM')])
         water = ihm.Entity([ihm.WaterChemComp()])
+        sugar = ihm.Entity([ihm.SaccharideChemComp('NAG'),
+                            ihm.SaccharideChemComp('FUC')])
         self.assertEqual(protein.type, 'polymer')
         self.assertTrue(protein.is_polymeric())
+        self.assertFalse(protein.is_branched())
         self.assertEqual(heme.type, 'non-polymer')
         self.assertFalse(heme.is_polymeric())
+        self.assertFalse(heme.is_branched())
         self.assertEqual(water.type, 'water')
         self.assertFalse(water.is_polymeric())
+        self.assertFalse(water.is_branched())
+        self.assertEqual(sugar.type, 'branched')
+        self.assertFalse(sugar.is_polymeric())
+        self.assertTrue(sugar.is_branched())
+
+        # A single sugar should be classified non-polymer
+        single_sugar = ihm.Entity([ihm.SaccharideChemComp('NAG')])
+        self.assertEqual(single_sugar.type, 'non-polymer')
+        self.assertFalse(single_sugar.is_polymeric())
+        self.assertFalse(single_sugar.is_branched())
+
+        # A single amino acid should be classified non-polymer
+        single_aa = ihm.Entity('A')
+        self.assertEqual(single_aa.type, 'non-polymer')
+        self.assertFalse(single_aa.is_polymeric())
+        self.assertFalse(single_aa.is_branched())
+
+        # ... unless forced polymer
+        single_aa._force_polymer = True
+        self.assertEqual(single_aa.type, 'polymer')
+        self.assertTrue(single_aa.is_polymeric())
+        self.assertFalse(single_aa.is_branched())
+
+        # An entity with no sequence is a polymer
+        empty = ihm.Entity([])
+        self.assertEqual(empty.type, 'polymer')
+        self.assertTrue(empty.is_polymeric())
+        self.assertFalse(empty.is_branched())
+
+        # ... unless hint branched
+        empty._hint_branched = True
+        self.assertEqual(empty.type, 'branched')
+        self.assertFalse(empty.is_polymeric())
+        self.assertTrue(empty.is_branched())
 
     def test_entity_src_method_default(self):
         """Test default values of Entity.src_method"""
@@ -210,9 +302,11 @@ class Tests(unittest.TestCase):
         water = ihm.Entity([ihm.WaterChemComp()])
         self.assertEqual(protein.src_method, "man")
         self.assertEqual(water.src_method, "nat")
+
         # src_method is readonly
         def try_set():
             protein.src_method = 'foo'
+
         self.assertRaises(TypeError, try_set)
 
     def test_entity_source(self):
@@ -236,16 +330,22 @@ class Tests(unittest.TestCase):
                           classification='7', description='8', location='9')
         s4 = ihm.Software(name='bar', version='1.0',
                           classification='a', description='b', location='c')
+        s5 = ihm.Software(name='bar', version=ihm.unknown,
+                          classification='a', description='b', location='c',
+                          citation='foo')
         # Should compare equal iff name and version both match
         self.assertEqual(s1, s3)
         self.assertEqual(hash(s1), hash(s3))
         self.assertNotEqual(s1, s2)
         self.assertNotEqual(s1, s4)
+        # Unknown values should not compare equal to known
+        self.assertNotEqual(s4, s5)
+        self.assertNotEqual(hash(s4), hash(s5))
 
     def test_citation(self):
         """Test Citation class"""
         s = ihm.Citation(title='Test paper', journal='J Mol Biol',
-                         volume=45, page_range=(1,20), year=2016,
+                         volume=45, page_range=(1, 20), year=2016,
                          authors=['Smith A', 'Jones B'],
                          doi='10.2345/S1384107697000225',
                          pmid='1234')
@@ -268,17 +368,18 @@ class Tests(unittest.TestCase):
     def test_citation_from_pubmed_id(self):
         """Test Citation.from_pubmed_id()"""
         c = self._get_from_pubmed_id('pubmed_api.json')
-        self.assertEqual(c.title,
-                'Integrative structure and functional anatomy of a nuclear '
-                'pore complex (test of python-ihm lib).')
+        self.assertEqual(
+            c.title,
+            'Integrative structure and functional anatomy of a nuclear '
+            'pore complex (test of python-ihm lib).')
         self.assertEqual(c.journal, 'Nature')
         self.assertEqual(c.volume, '555')
-        self.assertEqual(c.page_range, ['475','482'])
+        self.assertEqual(c.page_range, ['475', '482'])
         self.assertEqual(c.year, '2018')
         self.assertEqual(c.pmid, 29539637)
         self.assertEqual(c.doi, '10.1038/nature26003')
         self.assertEqual(len(c.authors), 32)
-        self.assertEqual(c.authors[0], 'Kim SJ')
+        self.assertEqual(c.authors[0], 'Kim, S.J.')
 
     def test_citation_from_pubmed_id_one_page(self):
         """Test Citation.from_pubmed_id() with page rather than range"""
@@ -294,12 +395,13 @@ class Tests(unittest.TestCase):
     def test_citation_from_pubmed_id_no_doi(self):
         """Test Citation.from_pubmed_id() with no DOI"""
         c = self._get_from_pubmed_id('pubmed_api_no_doi.json')
-        self.assertEqual(c.title,
-                'Integrative structure and functional anatomy of a nuclear '
-                'pore complex (test of python-ihm lib).')
+        self.assertEqual(
+            c.title,
+            'Integrative structure and functional anatomy of a nuclear '
+            'pore complex (test of python-ihm lib).')
         self.assertIsNone(c.doi)
         # Make sure that page range "475-82" is handled as 475,482
-        self.assertEqual(c.page_range, ['475','482'])
+        self.assertEqual(c.page_range, ['475', '482'])
 
     def test_entity_residue(self):
         """Test Residue derived from an Entity"""
@@ -308,16 +410,36 @@ class Tests(unittest.TestCase):
         self.assertEqual(r.entity, e)
         self.assertIsNone(r.asym)
         self.assertEqual(r.seq_id, 3)
+        self.assertEqual(r.comp.id, 'CYS')
+
+    def test_water_asym(self):
+        """Test WaterAsymUnit class"""
+        e = ihm.Entity('AHCDAH')
+        water = ihm.Entity([ihm.WaterChemComp()])
+        a = ihm.AsymUnit(e)
+        self.assertEqual(a.seq_id_range, (1, 6))
+        self.assertEqual(len(a.sequence), 6)
+        self.assertEqual(a.number_of_molecules, 1)
+
+        a = ihm.WaterAsymUnit(water, number=3)
+        self.assertEqual(a.seq_id_range, (1, 3))
+        self.assertEqual(len(a.sequence), 3)
+        self.assertEqual(a.number_of_molecules, 3)
+
+        self.assertRaises(TypeError, ihm.AsymUnit, water)
+        self.assertRaises(TypeError, ihm.WaterAsymUnit, e)
 
     def test_asym_unit_residue(self):
         """Test Residue derived from an AsymUnit"""
         e = ihm.Entity('AHCDAH')
         a = ihm.AsymUnit(e, auth_seq_id_map=5)
         r = a.residue(3)
-        self.assertIsNone(r.entity)
+        self.assertEqual(r.entity, e)
         self.assertEqual(r.asym, a)
         self.assertEqual(r.seq_id, 3)
         self.assertEqual(r.auth_seq_id, 8)
+        self.assertIsNone(r.ins_code)
+        self.assertEqual(r.comp.id, 'CYS')
 
     def test_atom_entity(self):
         """Test Atom class built from an Entity"""
@@ -337,7 +459,7 @@ class Tests(unittest.TestCase):
         a = asym.residue(3).atom('CA')
         self.assertEqual(a.id, 'CA')
         self.assertEqual(a.residue.seq_id, 3)
-        self.assertIsNone(a.entity)
+        self.assertEqual(a.entity, e)
         self.assertEqual(a.asym, asym)
         self.assertEqual(a.seq_id, 3)
 
@@ -345,66 +467,121 @@ class Tests(unittest.TestCase):
         """Test EntityRange class"""
         e = ihm.Entity('AHCDAH')
         heme = ihm.Entity([ihm.NonPolymerChemComp('HEM')])
+        sugar = ihm.Entity([ihm.SaccharideChemComp('NAG')])
         e._id = 42
-        self.assertEqual(e.seq_id_range, (1,6))
-        r = e(3,4)
-        self.assertEqual(r.seq_id_range, (3,4))
+        self.assertEqual(e.seq_id_range, (1, 6))
+        r = e(3, 4)
+        self.assertEqual(r.seq_id_range, (3, 4))
         self.assertEqual(r._id, 42)
-        # Cannot create ranges for nonpolymeric entities
+        # Cannot create ranges for nonpolymeric or branched entities
         self.assertRaises(TypeError, heme.__call__, 1, 1)
-        samer = e(3,4)
-        otherr = e(2,4)
+        self.assertRaises(TypeError, sugar.__call__, 1, 1)
+        samer = e(3, 4)
+        otherr = e(2, 4)
         self.assertEqual(r, samer)
         self.assertEqual(hash(r), hash(samer))
         self.assertNotEqual(r, otherr)
-        self.assertNotEqual(r, e) # entity_range != entity
+        self.assertNotEqual(r, e)  # entity_range != entity
 
     def test_asym_range(self):
         """Test AsymUnitRange class"""
         e = ihm.Entity('AHCDAH')
         heme = ihm.Entity([ihm.NonPolymerChemComp('HEM')])
-        a = ihm.AsymUnit(e)
+        sugar = ihm.Entity([ihm.SaccharideChemComp('NAG'),
+                            ihm.SaccharideChemComp('FUC')])
+        a = ihm.AsymUnit(e, "testdetail")
         aheme = ihm.AsymUnit(heme)
+        asugar = ihm.AsymUnit(sugar)
         a._id = 42
-        self.assertEqual(a.seq_id_range, (1,6))
+        self.assertEqual(a.seq_id_range, (1, 6))
         # seq_id is not defined for nonpolymers
-        self.assertEqual(aheme.seq_id_range, (None,None))
-        r = a(3,4)
-        self.assertEqual(r.seq_id_range, (3,4))
+        self.assertEqual(aheme.seq_id_range, (None, None))
+        # We use seq_id internally for branched entities
+        self.assertEqual(asugar.seq_id_range, (1, 2))
+        r = a(3, 4)
+        self.assertEqual(r.seq_id_range, (3, 4))
         self.assertEqual(r._id, 42)
         self.assertEqual(r.entity, e)
-        # Cannot create ranges for nonpolymeric entities
+        self.assertEqual(r.details, "testdetail")
+        # Cannot create ranges for nonpolymeric or branched entities
         self.assertRaises(TypeError, aheme.__call__, 1, 1)
-        samer = a(3,4)
-        otherr = a(2,4)
+        self.assertRaises(TypeError, asugar.__call__, 1, 1)
+        samer = a(3, 4)
+        otherr = a(2, 4)
         self.assertEqual(r, samer)
         self.assertEqual(hash(r), hash(samer))
         self.assertNotEqual(r, otherr)
-        self.assertNotEqual(r, a)      # asym_range != asym
-        self.assertNotEqual(r, e(3,4)) # asym_range != entity_range
-        self.assertNotEqual(r, e)      # asym_range != entity
+        self.assertNotEqual(r, a)        # asym_range != asym
+        self.assertNotEqual(r, e(3, 4))  # asym_range != entity_range
+        self.assertNotEqual(r, e)        # asym_range != entity
+
+    def test_asym_segment(self):
+        """Test AsymUnitSegment class"""
+        e = ihm.Entity('AHCDAH')
+        a = ihm.AsymUnit(e)
+        seg = a.segment('AH--CD', 1, 4)
+        self.assertEqual(seg.gapped_sequence, 'AH--CD')
+        self.assertEqual(seg.seq_id_range, (1, 4))
 
     def test_auth_seq_id_offset(self):
         """Test auth_seq_id offset from seq_id"""
         e = ihm.Entity('AHCDAH')
         a = ihm.AsymUnit(e, auth_seq_id_map=5)
-        self.assertEqual(a._get_auth_seq_id(1), 6)
+        self.assertEqual(a._get_auth_seq_id_ins_code(1), (6, None))
 
     def test_auth_seq_id_dict(self):
         """Test auth_seq_id dict map from seq_id"""
         e = ihm.Entity('AHCDAH')
-        a = ihm.AsymUnit(e, auth_seq_id_map={1:0, 2:4})
-        self.assertEqual(a._get_auth_seq_id(1), 0)
-        self.assertEqual(a._get_auth_seq_id(2), 4)
-        self.assertEqual(a._get_auth_seq_id(3), 3)
+        a = ihm.AsymUnit(e, auth_seq_id_map={1: 0, 2: (4, 'A')})
+        self.assertEqual(a._get_auth_seq_id_ins_code(1), (0, None))
+        self.assertEqual(a._get_auth_seq_id_ins_code(2), (4, 'A'))
+        self.assertEqual(a._get_auth_seq_id_ins_code(3), (3, None))
 
     def test_auth_seq_id_list(self):
         """Test auth_seq_id list map from seq_id"""
         e = ihm.Entity('AHCDAH')
         a = ihm.AsymUnit(e, auth_seq_id_map=[None, 0, 4])
-        self.assertEqual(a._get_auth_seq_id(1), 0)
-        self.assertEqual(a._get_auth_seq_id(2), 4)
-        self.assertEqual(a._get_auth_seq_id(3), 3)
+        self.assertEqual(a._get_auth_seq_id_ins_code(1), (0, None))
+        self.assertEqual(a._get_auth_seq_id_ins_code(2), (4, None))
+        self.assertEqual(a._get_auth_seq_id_ins_code(3), (3, None))
+
+    def test_orig_auth_seq_id_none(self):
+        """Test default orig_auth_seq_idm_map (None)"""
+        e = ihm.Entity('AHCDAH')
+        a = ihm.AsymUnit(e, auth_seq_id_map={1: 0, 2: (4, 'A')})
+        self.assertIsNone(a.orig_auth_seq_id_map)
+        self.assertEqual(a._get_pdb_auth_seq_id_ins_code(1), (0, 0, None))
+        self.assertEqual(a._get_pdb_auth_seq_id_ins_code(2), (4, 4, 'A'))
+        self.assertEqual(a._get_pdb_auth_seq_id_ins_code(3), (3, 3, None))
+
+    def test_orig_auth_seq_id_dict(self):
+        """Test orig_auth_seq_id_map as dict"""
+        e = ihm.Entity('AHCDAH')
+        a = ihm.AsymUnit(e, auth_seq_id_map={1: 0, 2: (4, 'A')},
+                         orig_auth_seq_id_map={1: 5})
+        self.assertEqual(a._get_pdb_auth_seq_id_ins_code(1), (0, 5, None))
+        self.assertEqual(a._get_pdb_auth_seq_id_ins_code(2), (4, 4, 'A'))
+        self.assertEqual(a._get_pdb_auth_seq_id_ins_code(3), (3, 3, None))
+
+    def test_water_orig_auth_seq_id_none(self):
+        """Test default water orig_auth_seq_id_map (None)"""
+        water = ihm.Entity([ihm.WaterChemComp()])
+        a = ihm.WaterAsymUnit(water, number=3,
+                              auth_seq_id_map={1: 0, 2: (4, 'A')})
+        self.assertIsNone(a.orig_auth_seq_id_map)
+        self.assertEqual(a._get_pdb_auth_seq_id_ins_code(1), (0, 0, None))
+        self.assertEqual(a._get_pdb_auth_seq_id_ins_code(2), (4, 4, 'A'))
+        self.assertEqual(a._get_pdb_auth_seq_id_ins_code(3), (3, 3, None))
+
+    def test_water_orig_auth_seq_id_dict(self):
+        """Test water orig_auth_seq_id_map as dict"""
+        water = ihm.Entity([ihm.WaterChemComp()])
+        a = ihm.WaterAsymUnit(water, number=3,
+                              auth_seq_id_map={1: 0, 2: (4, 'A')},
+                              orig_auth_seq_id_map={1: 5})
+        self.assertEqual(a._get_pdb_auth_seq_id_ins_code(1), (0, 5, None))
+        self.assertEqual(a._get_pdb_auth_seq_id_ins_code(2), (4, 4, 'A'))
+        self.assertEqual(a._get_pdb_auth_seq_id_ins_code(3), (3, 3, None))
 
     def test_assembly(self):
         """Test Assembly class"""
@@ -449,6 +626,13 @@ class Tests(unittest.TestCase):
                                   model_group=model_group4)
         e1.subsamples.extend((ss1, ss2))
         s.ensembles.append(e1)
+        # Ensemble without a model_group
+        e2 = ihm.model.Ensemble(model_group=None, num_models=10,
+                                post_process=None, name='cluster1',
+                                clustering_method='Hierarchical',
+                                clustering_feature='RMSD',
+                                precision=4.2)
+        s.ensembles.append(e2)
 
         mg = s._all_model_groups()
         # List may contain duplicates but only includes states
@@ -462,6 +646,30 @@ class Tests(unittest.TestCase):
                                     model_group3, model_group4,
                                     model_group1, model_group2])
 
+        model_group5 = 'mg5'
+        model_group6 = 'mg6'
+        model_group7 = 'mg7'
+        state3 = [model_group5]
+        state4 = [model_group6]
+        state5 = [model_group7]
+        mssc1 = ihm.multi_state_scheme.Connectivity(
+            begin_state=state3,
+            end_state=state4)
+        mssc2 = ihm.multi_state_scheme.Connectivity(
+            begin_state=state5)
+        mssc3 = ihm.multi_state_scheme.Connectivity(
+            begin_state=state3,
+            end_state=state4)
+        mss = ihm.multi_state_scheme.MultiStateScheme(
+            name='mss',
+            connectivities=[mssc1, mssc2, mssc3])
+        s.multi_state_schemes.append(mss)
+        mg = s._all_model_groups()
+        self.assertEqual(list(mg), [model_group1, model_group2,
+                                    model_group2, model_group2,
+                                    model_group5, model_group6,
+                                    model_group7])
+
     def test_all_models(self):
         """Test _all_models() method"""
         class MockModel(object):
@@ -558,26 +766,40 @@ class Tests(unittest.TestCase):
             pass
 
         c1 = ihm.Citation(title='Test paper', journal='J Mol Biol',
-                          volume=45, page_range=(1,20), year=2016,
+                          volume=45, page_range=(1, 20), year=2016,
                           authors=['Smith A', 'Jones B'],
                           doi='10.2345/S1384107697000225',
                           pmid='1234')
         c2 = ihm.Citation(title='Test paper', journal='J Mol Biol',
-                          volume=45, page_range=(1,20), year=2016,
+                          volume=45, page_range=(1, 20), year=2016,
                           authors=['Smith A', 'Jones B'],
                           doi='1.2.3.4',
                           pmid='1234')
-        rsr1 = MockObject() # Not a 3dem restraint
-        rsr2 = MockObject() # 3dem but with no provided citation
+        c3 = ihm.Citation(title='Test paper2', journal='J Mol Biol2',
+                          volume=46, page_range=(1, 20), year=2017,
+                          authors=['Smith A', 'Jones B'],
+                          doi='5.6.7.8',
+                          pmid='5678')
+        rsr1 = MockObject()  # Not a 3dem restraint
+        rsr2 = MockObject()  # 3dem but with no provided citation
         rsr2.fitting_method_citation_id = None
         rsr3 = MockObject()
         rsr2.fitting_method_citation_id = c1
 
+        s1 = ihm.Software(name='test', classification='test code',
+                          description='Some test program',
+                          version=1, location='http://test.org')
+        s2 = ihm.Software(name='test', classification='test code',
+                          description='Some test program',
+                          version=1, location='http://test.org',
+                          citation=c3)
+
         s = ihm.System()
         s.restraints.extend((rsr1, rsr2, rsr3))
         s.citations.extend((c2, c2))
+        s.software.extend((s1, s2))
         # duplicates should be filtered globally
-        self.assertEqual(list(s._all_citations()), [c2, c1])
+        self.assertEqual(list(s._all_citations()), [c2, c3, c1])
 
     def test_all_software(self):
         """Test _all_software() method"""
@@ -652,6 +874,7 @@ class Tests(unittest.TestCase):
         """Test _all_locations() method"""
         class MockObject(object):
             pass
+
         class MockDataset(object):
             parents = []
         loc1 = MockObject()
@@ -720,6 +943,7 @@ class Tests(unittest.TestCase):
         """Test _all_datasets() method"""
         class MockObject(object):
             pass
+
         class MockDataset(object):
             parents = []
 
@@ -859,7 +1083,6 @@ class Tests(unittest.TestCase):
         # duplicates should not be filtered
         self.assertEqual(list(s._all_pseudo_sites()), [s1, s2, s2, s1, s2])
 
-
     def test_all_chem_descriptors(self):
         """Test _all_chem_descriptors() method"""
         class MockObject(object):
@@ -875,8 +1098,6 @@ class Tests(unittest.TestCase):
         s.flr_data.append(f)
         r1 = MockObject()
         r2 = MockObject()
-        r2.linker = None
-        r3 = MockObject()
         r2.linker = d3
         s.restraints.extend((r1, r2))
 
@@ -884,8 +1105,9 @@ class Tests(unittest.TestCase):
         s.orphan_chem_descriptors.extend((d1, d2, d1))
 
         # FLR chemical descriptors
-        conj = ihm.flr.PolyProbeConjugate(sample_probe=None, chem_descriptor=d4,
-                                          ambiguous_stoichiometry=False)
+        conj = ihm.flr.PolyProbeConjugate(
+            sample_probe=None, chem_descriptor=d4,
+            ambiguous_stoichiometry=False)
         f.poly_probe_conjugates.append(conj)
 
         # duplicates should not be filtered
@@ -901,8 +1123,8 @@ class Tests(unittest.TestCase):
         a1 = ihm.AsymUnit(e1)
         s.entities.append(e1)
         s.asym_units.append(a1)
-        e1rng = e1(1,3)
-        a1rng = a1(1,2)
+        e1rng = e1(1, 3)
+        a1rng = a1(1, 2)
 
         sm1 = MockObject()
         sm1.asym_unit = e1rng
@@ -928,6 +1150,164 @@ class Tests(unittest.TestCase):
         self.assertEqual(list(s._all_entity_ranges()),
                          [e1rng, a1, e1, a1, a1rng])
 
+    def test_all_multi_state_schemes(self):
+        """Test _all_multi_state_schemes() method"""
+        class MockObject(object):
+            pass
+
+        s = ihm.System()
+        m1 = MockObject()
+        m2 = MockObject()
+        m3 = MockObject()
+        s.multi_state_schemes.append(m1)
+        s.multi_state_schemes.append(m2)
+        s.multi_state_schemes.append(m3)
+
+        self.assertEqual(list(s._all_multi_state_schemes()),
+                         [m1, m2, m3])
+
+    def test_all_multi_state_scheme_connectivities(self):
+        """Test _all_multi_state_scheme_connectivities() method"""
+        class MockObject(object):
+            pass
+
+        s = ihm.System()
+        # Multi-state schemes
+        mss1 = ihm.multi_state_scheme.MultiStateScheme('mss1')
+        mss2 = ihm.multi_state_scheme.MultiStateScheme('mss2')
+        # States
+        s1 = MockObject()
+        s2 = MockObject()
+        s3 = MockObject()
+        # Connectivities
+        c1 = ihm.multi_state_scheme.Connectivity(s1)
+        c2 = ihm.multi_state_scheme.Connectivity(s2)
+        c3 = ihm.multi_state_scheme.Connectivity(s3)
+        mss1.add_connectivity(c1)
+        mss1.add_connectivity(c2)
+        mss2.add_connectivity(c1)
+        mss2.add_connectivity(c3)
+        s.multi_state_schemes.append(mss1)
+        s.multi_state_schemes.append(mss2)
+        # Duplicates are kept
+        self.assertEqual(list(s._all_multi_state_scheme_connectivities()),
+                         [c1, c2, c1, c3])
+
+    def test_all_kinetic_rates(self):
+        """Test _all_kinetic_rates() method"""
+        class MockObject(object):
+            pass
+
+        s = ihm.System()
+        # Multi-state schemes
+        mss1 = ihm.multi_state_scheme.MultiStateScheme('mss1')
+        mss2 = ihm.multi_state_scheme.MultiStateScheme('mss2')
+        # States
+        s1 = MockObject()
+        s2 = MockObject()
+        s3 = MockObject()
+        # Kinetic rates
+        k1 = MockObject()
+        k2 = MockObject()
+        k3 = MockObject()
+
+        # Connectivities
+        c1 = ihm.multi_state_scheme.Connectivity(
+            begin_state=s1,
+            kinetic_rate=k1)
+        c2 = ihm.multi_state_scheme.Connectivity(
+            begin_state=s2,
+            kinetic_rate=k2)
+        c3 = ihm.multi_state_scheme.Connectivity(
+            begin_state=s3,
+            kinetic_rate=k3)
+        c4 = ihm.multi_state_scheme.Connectivity(
+            begin_state=s3,
+            kinetic_rate=None)
+
+        mss1.add_connectivity(c1)
+        mss1.add_connectivity(c2)
+        mss2.add_connectivity(c1)
+        mss2.add_connectivity(c3)
+        mss2.add_connectivity(c4)
+        s.multi_state_schemes.append(mss1)
+        s.multi_state_schemes.append(mss2)
+
+        # Does not contain duplicates
+        self.assertEqual(list(s._all_kinetic_rates()),
+                         [k1, k2, k3])
+
+        # From kinetic_rate_fret_analysis_connections in FLRData
+        k4 = MockObject()
+        kfc = MockObject()
+        kfc.kinetic_rate = k4
+        f = ihm.flr.FLRData()
+        f.kinetic_rate_fret_analysis_connections.append(kfc)
+        s.flr_data.append(f)
+        # Does not contain duplicates
+        self.assertEqual(list(s._all_kinetic_rates()),
+                         [k1, k2, k3, k4])
+
+    def test_all_relaxation_times(self):
+        """Test _all_relaxation_times() method"""
+        class MockObject(object):
+            pass
+        s = ihm.System()
+        # Multi-state schemes
+        mss1 = ihm.multi_state_scheme.MultiStateScheme('mss1')
+        mss2 = ihm.multi_state_scheme.MultiStateScheme('mss2')
+        # States
+        s1 = MockObject()
+        s2 = MockObject()
+        s3 = MockObject()
+        # Kinetic rates
+        r1 = MockObject()
+        r2 = MockObject()
+        r3 = MockObject()
+        r4 = MockObject()
+        r5 = MockObject()
+
+        # Connectivities
+        c1 = ihm.multi_state_scheme.Connectivity(
+            begin_state=s1,
+            relaxation_time=r1)
+        c2 = ihm.multi_state_scheme.Connectivity(
+            begin_state=s2,
+            relaxation_time=r2)
+        c3 = ihm.multi_state_scheme.Connectivity(
+            begin_state=s3,
+            relaxation_time=r3)
+        c4 = ihm.multi_state_scheme.Connectivity(
+            begin_state=s3,
+            relaxation_time=None,
+            kinetic_rate='rate'
+        )
+
+        mss1.add_relaxation_time(r4)
+        mss2.add_relaxation_time(r5)
+
+        mss1.add_connectivity(c1)
+        mss1.add_connectivity(c2)
+        mss2.add_connectivity(c1)
+        mss2.add_connectivity(c3)
+        mss2.add_connectivity(c4)
+        s.multi_state_schemes.append(mss1)
+        s.multi_state_schemes.append(mss2)
+        # Does not contain duplicates
+        self.assertEqual(list(s._all_relaxation_times()),
+                         [r4, r5, r1, r2, r3])
+
+        # From relaxation_time_fret_analysis_connections in FLRData
+        r6 = MockObject()
+        rfc = MockObject()
+        rfc.relaxation_time = r6
+        f = ihm.flr.FLRData()
+        f.relaxation_time_fret_analysis_connections.append(rfc)
+        s.flr_data.append(f)
+        # Does not contain duplicates
+        self.assertEqual(list(s._all_relaxation_times()),
+                         [r4, r5, r1, r2, r3, r6])
+
     def test_update_locations_in_repositories(self):
         """Test update_locations_in_repositories() method"""
         s = ihm.System()
@@ -952,6 +1332,29 @@ class Tests(unittest.TestCase):
         # Should act like False
         self.assertFalse(u)
 
+    def test_branch_descriptor(self):
+        """Test the BranchDescriptor class"""
+        bd = ihm.BranchDescriptor(text='foo', type='bar', program='baz',
+                                  program_version="1.0")
+        self.assertEqual(bd.text, 'foo')
+        self.assertEqual(bd.type, 'bar')
+        self.assertEqual(bd.program, 'baz')
+        self.assertEqual(bd.program_version, '1.0')
+
+    def test_branch_link(self):
+        """Test the BranchLink class"""
+        lnk = ihm.BranchLink(num1=1, atom_id1='CA', leaving_atom_id1='H1',
+                             num2=2, atom_id2='N', leaving_atom_id2='H2',
+                             order='sing', details='foo')
+        self.assertEqual(lnk.num1, 1)
+        self.assertEqual(lnk.atom_id1, 'CA')
+        self.assertEqual(lnk.leaving_atom_id1, 'H1')
+        self.assertEqual(lnk.num2, 2)
+        self.assertEqual(lnk.atom_id2, 'N')
+        self.assertEqual(lnk.leaving_atom_id2, 'H2')
+        self.assertEqual(lnk.order, 'sing')
+        self.assertEqual(lnk.details, 'foo')
+
 
 if __name__ == '__main__':
     unittest.main()
